/*
  vt1xx terminal emulation by
  Rene Rebe <rene@exactcode.de>
  
  Lossy based on code of bogl-term by:
  Edmund GRIMLEY EVANS <edmundo@rano.org>
  
  Additional information from the Linux kernel driver/char/vt.c by:
  Linus Torvalds
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA. */

#include "eTerm.hh"

#include <iostream>

#include "X11Helper.hh"
#include "EvasHelper.hh"

static const unsigned char palette[16][3] = 
  {
    {255, 255, 255},	/* 7: High wight */
    {255, 0,   0},	/* 1: Light red. */
    {0,   255, 0},	/* 2: Light green */
    {255, 255, 0},	/* 3: Yellow. */
    {0,   0,   255},	/* 4: Light blue */
    {255, 0,   255},	/* 5: Light magenta */
    {0,   255, 255},	/* 6: Light cyan */
    {0,   0,   0},	/* 0: Black. */
    {128, 128, 128},	/* 8: Gray */
    {128, 0,   0},	/* 9: Red */
    {0,   128, 0},	/* A: Green */
    {128, 128, 0},	/* B: Brown */
    {0,   0,   128},	/* C: Blue */
    {128, 0,   128},	/* D: Magenta */
    {0,   128, 128},	/* E: Cyan */
    {192, 192, 192},	/* F: White */
   };

// display stuff

eTerm::eTerm (EvasCanvas* _evas, int _xsize, int _ysize)
  : Term (_xsize, _ysize)
{
  evas = _evas;
  escreen = 0;
  escreen_back = 0;
  Resize (_xsize, _ysize);
}

void eTerm::Resize (int _xsize, int _ysize)
{
  // deallocate ?
  if (escreen || escreen_back)
    {
      std::cerr << "deallocateding evas ..." << std::endl;
      for (int y = 0; y < ysize; ++y)
	for (int x = 0; x < xsize; ++x) {
	  delete escreen [y*xsize + x];
	  delete escreen_back [y*xsize + x];
	}
      free (escreen); escreen = 0;
      free (escreen_back); escreen_back = 0;
    }
  
  std::cerr << "allocating new ..." << std::endl;
  escreen      = (EvasText**) malloc (sizeof(EvasText**) * _xsize * _ysize);
  escreen_back = (EvasRectangle**) malloc (sizeof(EvasRectangle**) * _xsize * _ysize);
  
  // resize underlying terminal
  resize (_xsize, _ysize);
  
  for (int y = 0; y < ysize; ++y)
    for (int x = 0; x < xsize; ++x)
      {
	escreen [y*_xsize + x] = new EvasText (*evas);
	escreen_back [y*_xsize + x] = new EvasRectangle (*evas);
	
	EvasText* text = escreen [y*_xsize + x];
	EvasRectangle* rect = escreen_back [y*_xsize + x];
	
	text->Layer (10);
	text->Font ("Vera Sans Mono", 12);
	if (x == 0 && y == 0)
	  {
	    text->Text ("M");
	    text->Show ();
	    fw = text->Width ();
	    fh = text->Height ();
	  }
	
	text->Text (" ");
	text->Move (x*fw, y*fh);
	text->Color (255, 255, 255, 0xff);
	text->Show ();
	
	rect->Layer (5);
	rect->Move (x*fw, y*fh);
	rect->Resize (fw, fh);
	rect->Color (0, 0, 0, 0xff);
	rect->Show ();
      }
}

void eTerm::put_char (int x, int y, wchar_t wc,
		      int fg, int bg, int ul)
{
  char buf[MB_LEN_MAX];
  int k, r, w;
    
  wctomb(0, 0);
  if ((k = wctomb(buf, wc)) == -1)
    return;
    
  if (true /*bogl_in_font (font, wc)*/)
    {
      char txt [MAX_CCHARS + 1];
	
#if 0
      std::cerr << "[" << x << "," << y << "]: "
		<< (int) wc << "(" << (char) wc << ")" << std::endl;
#endif
      
      EvasText* etext = escreen [y*xsize + x];
      EvasRectangle* erect = escreen_back [y*xsize + x];
      
      wctomb(0, 0);
      if ((k = wctomb (txt, wc)) != -1)
	{
	  txt [k] = 0;
	  std::string text = etext->GetText ();
	  std::string ntext (txt);
	  if (ntext != text) {
	    etext->Text (txt);
	    // TODO: this is a text balancing, prettification HACK!
	    //       should always be 0 with a good Monospace font ...
	    int bias = (fw - etext->Width ()) / 2;
	    etext->Move (x*fw + bias, y*fh);
	  }
	  int r, g, b, a;
	  etext->GetColor (&r, &g, &b, &a);
	    
	  int nr, ng, nb, na = 255;
	  nr = palette[fg][0];
	  ng = palette[fg][1];
	  nb = palette[fg][2];
	    
	  if (nr != r || ng != g || nb != b)
	    etext->Color (nr, ng, nb, na);
	  // TODO: ul - underline
	    
	  erect->GetColor (&r, &g, &b, &a);
	  nr = palette[bg][0];
	  ng = palette[bg][1];
	  nb = palette[bg][2];
	  if (nr != r || ng != g || nb != b)
	    erect->Color (nr, ng, nb, na);
	}
    }
  else
    {
      /* repeat the default char w times */
      w = wcwidth(wc);
      for (r = 0; r < w; r++)
	// TODO: bogl_text (XPOS(x + r), YPOS(y), buf, k, fg, bg, ul, font);
	;
    }
}

